#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

/* Maximum amount of link-chasing we will do. */
#define MAX_SYMLINK_LEVELS 50

#ifndef MAXFILELEN
#define MAXFILELEN 1024
#endif

#ifndef HAVE_MEMMOVE
#define memmove(a,b,n) bcopy(b,a,n)
#endif

char *prog;
int debug=0;
int terse=0;
int longlist=0;
int docomponents=0;

/*
 * Here is a goofy but functional way to decide if -g should be passed to
 * ls(1).
 */
char *
lsoptions() {
    FILE *fp;
    char buf[200];
    char *s;
    int n, ng;
    static char LSoptions[20] = "";

    if (!LSoptions[0]) {
	fp = popen("/bin/ls -ldg .", "r");
	s = fgets(buf, sizeof(buf), fp);
	pclose(fp);
	ng = strlen(s);

	fp = popen("/bin/ls -ld .", "r");
	s = fgets(buf, sizeof(buf), fp);
	pclose(fp);
	n = strlen(s);

	if (n >= ng)
	    strcpy(LSoptions, "-ld");
	else
	    strcpy(LSoptions, "-ldg");
    }

    return LSoptions;
}

void
usage()
{
    char *lsoptions();
    printf("%s v %s.%s\n", prog, VERSION, PATCHLEVEL);
    printf("Use: %s [-h] [-c] [-d] [-l] [-t] file...\n", prog);
    printf(
	"Purpose: follows symbolic links and prints the pointed-at names.\n");
    printf("Options: \n");
    printf("\t-h: prints this message, then exits\n");
    printf("\t-c: follows the components of the path, one at a time.\n");
    printf("\t-d: print debug info; not usually useful.\n");
    printf("\t-l: does ls %s on the pointed-at file\n", lsoptions());
    printf("\t-t: terse: print just expanded pointed-at pathnames\n");
}

int
main(int argc, char **argv)
{
    extern int optind;
    int getoptions(int, char **);
    int follow(char *);
    int firsttime;
    char *s, *file_end, *comp_end, *file;
    char *start_dir;

    prog = strrchr(argv[0], '/');
    prog = (prog) ? prog+1 : argv[0];

    if ((start_dir = getcwd(NULL, MAXFILELEN+1)) == NULL) {
	int e=errno;
	fprintf(stderr, "%s: can't get current directory: ", prog);
	errno = e;
	perror("");
	exit(1);
    }

    getoptions(argc, argv);
    for (firsttime=1; optind < argc; firsttime=0, optind++) {
	/* If we are doing multiple files, make sure we return to
	 * the start directory.
	 */
	if (!firsttime) {
	    if (chdir(start_dir) == -1) {
		int e=errno;
		fprintf(stderr, "%s: Failed to chdir to %s: ", prog, start_dir);
		errno = e;
		perror("");
		exit(1);
	    }
	}
	file = argv[optind];
	file_end = file + strlen(file);
	if (docomponents && file_end > file) {
	    for (s = strtok(file+1, "/"); s; s = strtok(NULL, "/")) {
		if (follow(file) == -1)
		    return;
		/* Restore the slash */
		comp_end = file + strlen(file);
		if (comp_end < file_end)
		    *comp_end = '/';
	    }
	} else {
	    follow(file);
	}
    }
}

int
getoptions(int argc, char **argv)
{
    extern int      getopt(int, char *const *, const char *);
    int c;
    while ((c = getopt(argc, argv, "cdhlt")) != -1) {
	switch (c) {
	case 'h':
	    usage();
	    exit(0);
	    break;
	case 'c':
	    docomponents = 1;
	    break;
	case 'd':
	    debug = 1;
	    break;
	case 'l':
	    longlist = 1;
	    break;
	case 't':
	    terse = 1;
	    break;
	default:
	    usage();
	    exit(1);
	    break;
	}
    }
    return 1;
}

int
follow(char *file)
{
    struct stat buf;
    char filebuf[2][MAXFILELEN+1];

    char dir[MAXFILELEN*2+1];
    char longWindyRoad[MAXFILELEN*2+1];
    struct {
	ino_t inode;
	dev_t devid;
    } beenthere[MAX_SYMLINK_LEVELS];

    char *s;
    int curbuf, l;
    int i, il, nlinks, self_ref;
    void puterr(int nlinks, char *fmt, char *arg);
    void putfile(int nlinks, char *file, char *dir, int arrow);

    strcpy(dir, ".");
    strcpy(longWindyRoad, ".");

    buf.st_mode = S_IFLNK;
    curbuf = 0;
    strcpy(filebuf[curbuf], file);

    for (nlinks=0, self_ref=0;
	nlinks < MAX_SYMLINK_LEVELS && !self_ref; nlinks++, curbuf=!curbuf) {

	if (debug) printf("------------\n");
	if (debug) printf("filebuf[%d]=<%s>\n", curbuf, filebuf[curbuf]);
	if (lstat(filebuf[curbuf], &buf) == -1) {
	    int e=errno;
	    putfile(nlinks, filebuf[curbuf], longWindyRoad, 0);
	    puterr(nlinks, "Failed to lstat %s: ", filebuf[curbuf]);
	    errno = e;
	    perror("");
	    return -1;
	}
	beenthere[nlinks].inode = buf.st_ino;
	beenthere[nlinks].devid = buf.st_dev;

	if (!S_ISLNK(buf.st_mode))
	    break;

	if ((l=readlink(filebuf[curbuf], filebuf[!curbuf],
				sizeof(filebuf[!curbuf])-1)) == -1) {
	    int e=errno;
	    putfile(nlinks, filebuf[curbuf], longWindyRoad, 0);
	    puterr(nlinks, "Failed to read link %s: ", filebuf[curbuf]);
	    errno = e;
	    perror("");
	    return -1;
	}
	filebuf[!curbuf][l] = '\0';
	if (debug) printf("filebuf[%d]=<%s>\n", !curbuf, filebuf[!curbuf]);

	if (!terse)
	    putfile(nlinks, filebuf[curbuf], longWindyRoad, 1);

	/* Chdir to the directory containing the file of interest.
	 * This is so that we can easily follow relative links.
	 */
	strcpy(dir, filebuf[curbuf]);
	if (debug) printf("dir=<%s>\n", dir);
	if ( *filebuf[curbuf] == '/' ) {
	    strcpy(longWindyRoad,filebuf[curbuf]);
	} else {
	    strcat(longWindyRoad,"/");
	    strcat(longWindyRoad,filebuf[curbuf]);
	}
	if (debug) printf("longWindyRoad=<%s>... ", longWindyRoad);
	if ((s = strrchr(longWindyRoad, '/'))) {
	    if (s == longWindyRoad)
		*(s+1) = '\0';
	    else
		*s = '\0';
	}
	if (debug) printf("<%s>\n", longWindyRoad);

	strcpy(dir, filebuf[curbuf]);
	if ((s = strrchr(dir, '/'))) {
	    if (s == dir)
		*(s+1) = '\0';
	    else
		*s = '\0';
	} else {
	    strcpy(dir, ".");
	}

	/* If longWindyRoad ends with XXX/.. (XXX != ".") eliminate the pair. */
	if ((s = strstr(longWindyRoad+1, "/..")) && (*(s+3) == '\0') &&
	    ( *(s-1) != '.' || ((s > (longWindyRoad+1)) && *(s-2) != '/'))) {
	    *s = '\0';
	    if ((s = strrchr(longWindyRoad, '/'))) {
		*s = '\0';
	    } else {
		strcpy(longWindyRoad, ".");
	    }
	}
	if (debug) printf("longWindyRoad=<%s>; dir=<%s>\n", longWindyRoad, dir);

	if (chdir(dir) == -1) {
	    int e=errno;
	    putfile(nlinks, filebuf[curbuf], longWindyRoad, 0);
	    puterr(nlinks, "Failed to chdir to %s: ", dir);
	    errno = e;
	    perror("");
	    return -1;
	}

	/* Test if symlink loop */
	for (il = nlinks-1; il >= 0; il--) {
	    if (beenthere[nlinks].devid == beenthere[il].devid &&
			beenthere[nlinks].inode == beenthere[il].inode) {
		/* Self-referential */
		self_ref = 1;
		break;
	    }
	}
    }
    if (self_ref) {
	if (!longlist || !terse)
	    putfile(nlinks, filebuf[curbuf], longWindyRoad, 0);
	fprintf(stderr, "Symbolic link loop.\n");
	return -1;
    } else if (S_ISLNK(buf.st_mode)) {
	fprintf(stderr, "Too many levels of symbolic links.\n");
	return -1;
    }
    putfile(nlinks, filebuf[curbuf], longWindyRoad, 0);
    if (longlist) {
	char buf[MAXFILELEN*3];
	sprintf(buf, "/bin/ls %s %s", lsoptions(), filebuf[curbuf]);
	system(buf);
    }
    return 0;
}

void
indent(FILE *fp, int nlinks)
{
    int i;
    for (i=0; i<nlinks; i++)
	fputs("  ", fp);
}

void
puterr(int nlinks, char *fmt, char *arg)
{
    indent(stderr, nlinks);
    fprintf(stderr, fmt, arg);
}

void
putfile(int nlinks, char *file, char *dir, int arrow)
{
    char *s;
    int m, endslash;
    extern int terse;

    if (!terse) {
	indent(stdout, nlinks);
	fputs(file, stdout);
    }

    m = strlen(dir);
    endslash = (m && dir[m-1] == '/');

    if (debug) fprintf(stderr, " [[nlinks=%d dir=<%s> file=<%s> arrow=%d]] ",
				nlinks, dir, file, arrow);
    if (file[0] == '/')
	printf("%s\n", arrow ? " ->" : "");
    else {
	/* If file begins "../", and dir ends with /XXX, eliminate the pair.
	 * If file begins "./", eliminate it.
	 * If dir ends with "/", don't print extra "/" separator.
	 */
	if ((strncmp(file, "../", 3) == 0) && (s=strrchr(dir, '/')) && *(s+1)) {
	    printf("%s%.*s%s%s\n", !terse ? "  ( " : "",
			    (s-dir), dir, endslash ? "" : "/", file+3,
			    (arrow && !terse) ? " ) ->" : !terse ? " )" : "");
	} else if (strncmp(file, "./", 2) == 0) {
	    printf("%s%.*s%s%s%s\n", !terse ? "  ( " : "",
			    (s-dir), dir, endslash ? "" : "/", file+2,
			    (arrow && !terse) ? " ) ->" : !terse ? " )" : "");
	} else {
	    printf("%s%s%s%s%s\n", !terse ? "  ( " : "",
			    dir, endslash ? "" : "/", file,
			    (arrow && !terse) ? " ) ->" : !terse ? " )" : "");
	}
    }
}
